import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from nn2.etc import get_model_n_params, get_cnn_szs


def init_layer(l):
    if getattr(l, 'bias', None) is not None: nn.init.constant_(l.bias, 0) # l.bias.data.zero_()
    if isinstance(l, (nn.BatchNorm1d, nn.BatchNorm2d)): l.weight.data.fill_(1)
    if isinstance(l, (nn.Linear,)): nn.init.kaiming_normal_(l.weight)
    if isinstance(l, (nn.Embedding,)): l.weight.data.normal_(0., 2./math.sqrt(l.weight.data.size(1)))
        #sc = 2/(w.size(1)+1)
        #w.uniform_(-sc, sc)
    if isinstance(l, (nn.Conv1d, nn.Conv2d)): nn.init.kaiming_normal_(l.weight)


def init_net(m):
    init_layer(m)
    for l in m.children(): init_net(l)


def noop(x): return x
    

# def ReLU_Sim(x): return x.clamp_min(0.) - 0.5
class ReLU_Sim(nn.Module):
    def forward(self, x): return x.clamp_min(0.) - 0.5


class Mish(nn.Module):
    def __init__(self): super().__init__()
    def forward(self, x): return x *( torch.tanh(F.softplus(x)))


default_act =  Mish() # nn.ReLU(inplace=True), Mish()


class Sigmoid(nn.Module):
    def __init__(self, yrange=None):
        super().__init__()
        self.yrange = yrange
    def forward(self, x):
        return torch.sigmoid(x) if self.yrange is None else self.y_range[0]+(self.yrange[1]-self.yrange[0])*torch.sigmoid(x)

# def sigmoid(input, eps=1e-7):
#     "Same as `torch.sigmoid`, plus clamping to `(eps,1-eps)"
#     return input.sigmoid().clamp(eps,1-eps)


class Flatten(nn.Module):
    def forward(self, x): return x.view(x.size(0), -1)


def Linear(n1, n2, p=0.0, bn=True, act=default_act):
    net = []
    if bn:
        batch = nn.BatchNorm1d(num_features=n1)
        init_layer(batch)
        net += [batch]
    if p is not None: net += [nn.Dropout(p=p)]
    linear = nn.Linear(in_features=n1, out_features=n2)
    init_layer(linear)
    net += [linear]
    if act: net += [act]
    return nn.Sequential(*net)

def LinearFinal(n1, n2, p=None, act=nn.LogSoftmax(dim=1)): return Linear(n1, n2, p=p, bn=False, act=act)


def Conv2D(n1, n2, ks=3, stride=1, zero_bn=False, act=True):
    bn = nn.BatchNorm2d(n1)
    nn.init.constant_(bn.weight, 0. if zero_bn else 1.)
    conv = nn.Conv2d(n1, n2, kernel_size=ks, stride=stride, padding=ks//2, bias=False)
    layers = [bn, conv]
    if act: layers.append(default_act)
    return nn.Sequential(*layers)
    # PREVIOUS: return F.relu(self.batch(self.conv(x)))


def Conv1D(n1, n2, ks=3, stride=1, zero_bn=False, act=True):
    bn = nn.BatchNorm1d(n1)
    nn.init.constant_(bn.weight, 0. if zero_bn else 1.)
    conv = nn.Conv1d(n1, n2, kernel_size=ks, stride=stride, padding=ks//2, bias=False)
    layers = [bn, conv]
    if act: layers.append(default_act)
    return nn.Sequential(*layers)


class ResBlock2D(nn.Module):
    def __init__(self, ni, nf, stride=1):
        super().__init__()
        layers  = [Conv2D(ni, nf, stride=stride),
                   Conv2D(nf, nf, zero_bn=True, act=False)]
        self.convs = nn.Sequential(*layers)
        self.idconv = noop if ni==nf else Conv2D(ni, nf, ks=1, act=False)
        self.pool = noop if stride==1 else nn.AvgPool2d(2, ceil_mode=True)
    
    def forward(self, x): return default_act(self.convs(x) + self.idconv(self.pool(x)))


class ResBlock1D(nn.Module):
    def __init__(self, ni, nf, ks=3, stride=1):
        super().__init__()
        layers  = [Conv1D(ni, nf, ks=ks, stride=stride),
                   Conv1D(nf, nf, ks=ks, zero_bn=True, act=False)]
        self.convs = nn.Sequential(*layers)
        self.idconv = noop if ni==nf else Conv1D(ni, nf, ks=1, act=False)
        self.pool = noop if stride==1 else nn.AvgPool1d(2, ceil_mode=True)

    def forward(self, x): return default_act(self.convs(x) + self.idconv(self.pool(x)))


class TabularNet(nn.Module):
    def __init__(self, data, szs, emb_p, ps):
        super().__init__()
        self.ncats, self.nconts = data.ncats, data.nconts
        self.embs = nn.ModuleList([nn.Embedding(c, s) for c,s in data.emb_szs])
        self.nemb = sum(e.embedding_dim for e in self.embs)
        self.emb_drop = nn.Dropout(emb_p)
        self.bn_cont = nn.BatchNorm1d(data.nconts)
        szs = [self.nemb+data.nconts] + szs
        fcs = [Linear(szs[i], szs[i+1], p=ps[i]) for i in range(len(szs)-1)]
        out_sz = data.c if data.clsf else 1
        act_out = nn.LogSoftmax(dim=1) if data.clsf else Sigmoid()
        fcs += [LinearFinal(szs[-1], out_sz, act=act_out)]
        self.fcs = nn.Sequential(*fcs)
        init_net(self)
        np = sum([p.numel() for p in self.parameters()])
        i,c,e = data.ncats+data.nconts, len(data.emb_szs), self.nemb
        print(f'\nTabNet: {np:,d} params; {i:,d}->{i-c+e:,d} inputs; {c:,d}->{e:,d} embeds'.replace(',', '.'))

    def forward(self, x):
        x_cat, x_cont = x[:,:self.ncats].long(), x[:,self.ncats:].contiguous()
        if self.nemb!=0:
            x = [e(x_cat[:,i]) for i,e in enumerate(self.embs)]
            x = torch.cat(x, 1)
            x = self.emb_drop(x)
        if self.nconts!=0:
            # x_cont = self.bn_cont(x_cont)
            x = torch.cat([x, x_cont], 1) if self.nemb!=0 else x_cont
        x = self.fcs(x)
        return x


class SignalNet(nn.Module):
    def __init__(self, data, p=0.):
        super(SignalNet, self).__init__()
        x,_ = data.ds1[0]
        n1, sz = x.shape[0], x.shape[1]
        n2 = min((n1+1)*8,64)
        net = []
        szs = [(n1,sz)]
        for i in range(3):
            net += [Conv1D(n1, n2, stride=2)]
            n1, sz = n2, int(sz/2)
            szs += [(n1,sz)]
        bottle = int((int(math.log2(sz)))/2)
        for i in range(bottle):
            net += [ResBlock1D(n1, n1*2, ks=3, stride=2)]
            n1, sz = n1*2, int(sz/2)
            szs += [(n1,sz)]
            net += [nn.MaxPool1d(2)]  # AvgPool1d, MaxPool1d
            sz = int(sz/2)
            szs += [(n1,sz)]
        net += [nn.AdaptiveMaxPool1d(1)]
        net += [Flatten()]
        net += [Linear(1*n1, 1024, p=p)]
        net += [LinearFinal(1024, data.c, p=None)]
        self.net = nn.Sequential(*net)
        p = sum([p.numel() for p in self.parameters()])
        print(f'SignalNet: params: {p:,.0f}, signal_sz: {szs[0]}, bottles: {bottle}, last_conv {szs[-1]}')
        # print(szs)

    def forward(self, x): return self.net(x)


class Resnet(nn.Module):
    def __init__(self, data, layers=[64, 64, 64], p=None):
        super(Resnet, self).__init__()
        sz = [int(e) for e in data.ds1[0][0].shape]
        n1, n2 = sz[0], 32
        net = [ResBlock2D(n1, n2, stride=2)]
        n1 = n2
        for n2 in layers:
            net += [ResBlock2D(n1, n2, stride=2)]
            n1 = n2
        net += [nn.AdaptiveMaxPool2d(1)]
        net += [Flatten()]
        net += [Linear(layers[-1], 1024, p=p)]
        net += [LinearFinal(1024, data.c, p=None)]
        self.net = nn.Sequential(*net)
        p = sum([p.numel() for p in self.parameters()])
        self.szs = get_cnn_szs(self, sz)
        print(f'\nResnet: params: {p:,.0f}, img_sz: {sz}, last_conv {self.szs[-1]}: ')

    def forward(self, x): return self.net(x)









# class ResBlock(nn.Module):
#     def __init__(self, expansion, ni, nh, stride=1):
#         super().__init__()
#         nf,ni = nh*expansion,ni*expansion
#         layers  = [conv_layer(ni, nh, 3, stride=stride),
#                    conv_layer(nh, nf, 3, zero_bn=True, act=False)
#         ] if expansion == 1 else [
#                    conv_layer(ni, nh, 1),
#                    conv_layer(nh, nh, 3, stride=stride),
#                    conv_layer(nh, nf, 1, zero_bn=True, act=False)
#         ]
#         self.convs = nn.Sequential(*layers)
#         self.idconv = noop if ni==nf else conv_layer(ni, nf, 1, act=False)
#         self.pool = noop if stride==1 else nn.AvgPool2d(2, ceil_mode=True)

#     def forward(self, x): return act_fn(self.convs(x) + self.idconv(self.pool(x)))


# class XResNet(nn.Sequential):
#     @classmethod
#     def create(cls, expansion, layers, c_in=3, c_out=1000):
#         nfs = [c_in, (c_in+1)*8, 64, 64]
#         stem = [conv_layer(nfs[i], nfs[i+1], stride=2 if i==0 else 1)
#             for i in range(3)]

#         nfs = [64//expansion,64,128,256,512]
#         res_layers = [cls._make_layer(expansion, nfs[i], nfs[i+1], n_blocks=l, stride=1 if i==0 else 2)
#                   for i,l in enumerate(layers)]
#         res = cls(
#             *stem,
#             nn.MaxPool2d(kernel_size=3, stride=2, padding=1),
#             *res_layers,
#             nn.AdaptiveAvgPool2d(1), Flatten(),
#             nn.Linear(nfs[-1]*expansion, c_out),
#         )
#         init_net(res)
#         return res

#     @staticmethod
#     def _make_layer(expansion, ni, nf, n_blocks, stride):
#         return nn.Sequential(
#             *[ResBlock(expansion, ni if i==0 else nf, nf, stride if i==0 else 1)
#               for i in range(n_blocks)])


# def xresnet18 (**kwargs): return XResNet.create(1, [2, 2,  2, 2], **kwargs)
# def xresnet34 (**kwargs): return XResNet.create(1, [3, 4,  6, 3], **kwargs)
# def xresnet50 (**kwargs): return XResNet.create(4, [3, 4,  6, 3], **kwargs)
# def xresnet101(**kwargs): return XResNet.create(4, [3, 4, 23, 3], **kwargs)
# def xresnet152(**kwargs): return XResNet.create(4, [3, 8, 36, 3], **kwargs)



import random
import math
import itertools
from collections import defaultdict, Counter, OrderedDict
from collections.abc import Iterable, Iterator, Generator, Collection
import sys, datetime
from pathlib import Path
import numpy as np
import pandas as pd
import torch
from matplotlib import pyplot as plt
import seaborn as sns
from dlogicutils.core import *
from dlogicutils.ju import in_ipynb
T = torch.tensor

def get_sample(x1, y1, x2, y2, sz=1000, sz2=None):
    if sz2 is None: sz2=sz
    p = np.random.permutation(len(x1))
    x1, y1 = x1[p], y1[p]
    p = np.random.permutation(len(x2))
    x2, y2 = x2[p], y2[p]
    return x1[:sz], y1[:sz], x2[:sz2], y2[:sz2]

def plot(x, y):
    plt.figure()
    plt.plot(x,y)
    plt.show()
    plt.pause(.001)

def imgrid(x, y):
    if len(x.shape) not in [3,4]: Exception('x dims must be 3 ou 4')
    if len(x.shape)==3: x = x[None,:,:,:]
    n = x.shape[0]
    nc = 4
    nr = math.ceil(n/nc)
    fig, axs = plt.subplots(nr, nc, figsize=(15,4*nr))
    if type(axs)!=np.ndarray: axs = [[axs]]
    elif len(axs.shape)==1: axs = [axs]
    for i in range(n):
        r,c = i//nc,i%nc
        ax = axs[r][c]
        xi = x[i] if x[i].shape[2]>2 else x[i].squeeze(-1)
        ax.imshow(xi)
        ax.set_title(y[i])
    fig.tight_layout()
    plt.show()
    plt.pause(0.01)

def plot_signal(x, ch=0, title=None):
    plt.plot(x[ch])
    if title: plt.title(title)
    plt.show()
    plt.pause(0.001)
    
def sigshow(x, title):
    n = x.shape[0]
    nc = 2
    nr = math.ceil(n/nc)
    fig, axs = plt.subplots(nr, nc, figsize=(15,2*nr))
    for i in range(n):
        r,c = i//nc,i%nc
        axs[r][c].plot(x[i])
        axs[r][c].set_title(f'Channel {i}')
    fig.tight_layout()
    plt.show()
    plt.pause(0.001)


# def np_split(x, y, p=.2):
#     mask = np.random.permutation(len(x))
#     k = int(len(x)*(1-p))
#     x1, y1 = x[mask[:k]], y[mask[:k]]
#     x2, y2 = x[mask[k:]], y[mask[k:]]
#     return x1, y1, x2, y2


# def get_sample(x1, y1, x2, y2, sz=1000):
#     p = np.random.permutation(len(x1))
#     x1, y1 = x1[p], y1[p]
#     p = np.random.permutation(len(x2))
#     x2, y2 = x2[p], y2[p]
#     return x1[:sz], y1[:sz], x2[:sz], y2[:sz]
# x1, y1, x2, y2 = get_sample(x1, y1, x2, y2)
# x1.shape, y1.shape, x2.shape, y2.shape


# torch.Tensor.ndim = property(lambda x: len(x.shape))
def _tuple(x): return x if isinstance(x, (list, tuple)) else (x,x)
NoneType = type(None)
def noop (x=None, *args, **kwargs): return x
def t2s(t): return f'{int(t)//60:02d}:{int(t)%60:02d}'


def get_date_str(): 
    d = datetime.datetime.now()
    return  f'{d.year:04d}{d.month:02d}{d.day:02d}-{d.hour:02d}{d.minute:02d}{d.second:02d}'


def get_log_fname(path):
    dirname = path/'runs'
    fname = f'{dirname}/run-{get_date_str()}.csv'
    return fname, dirname


def save_model(model, path, stats):
    state = {'state_dict': model.state_dict(), 'stats': stats}
    torch.save(state, path)


def load_model(model, path):
    state = torch.load(path,  map_location='cpu')
    model.load_state_dict(state['state_dict'])
    return state['stats']


def get_model_n_params(model): return sum([p.numel() for p in model.parameters()])


def xy_shuffle_split(x, y, p=0.2):
    msk = np.random.permutation(len(x))
    n = int(len(x)*p)
    return x[msk[:n]], y[msk[:n]], x[msk[n:]], y[msk[n:]]


def get_cls_dist(y, cls_to_idx):
    if type(y)==list: y = np.array([cls_to_idx[e] for e in y])
    elif type(y)==torch.Tensor: y = y.numpy()
    c_idxs = [np.asarray(y==cls_to_idx[c]).nonzero()[0] for c in cls_to_idx]
    c_szs = [e.shape[0] for e in c_idxs]
    c_dist = [e.shape[0]/len(y) for e in c_idxs]
    return c_idxs, c_szs, c_dist


def get_cls_dist_string(cls, c_szs, c_dist):
    return ', '.join([f'({c} {s} {d:.2f})' for c,s,d in sorted(zip(cls, c_szs, c_dist), key = lambda t: -t[1])])


def get_cnn_szs(net, sz):
    if sz is None: return []
    C0, H0, W0 = sz
    C1, H1, W1 = C0, H0, W0
    szs = [(C0, H0, W0)]
    for name, l in net.named_modules():
        classname = l.__class__.__name__
        if classname == 'Conv2d':
            stride = _tuple(l.stride)
            padding = _tuple(l.padding)
            kernel_size = _tuple(l.kernel_size)
            dilation = _tuple(l.dilation)
            C1 = int(l.out_channels)
            H1 = int(math.floor((H0 + 2 * padding[0] - dilation[0] * (kernel_size[0] - 1) - 1) / stride[0] + 1))
            W1 = int(math.floor((W0 + 2 * padding[1] - dilation[1] * (kernel_size[1] - 1) - 1) / stride[1] + 1))
            szs.append([(C1, H1, W1)])
        C0, H0, W0 = C1, H1, W1
    return szs


def biased_roll(dist):
    r = random.random()
    cum, result = 0, 0
    for d in dist:
        cum += d
        if r < cum: return result
        result += 1


def table(cols, vals):
    pd.options.display.float_format = '{:,.4f}'.format
    display(pd.DataFrame(vals, columns=cols))


def show_dist(y):
    df = pd.DataFrame(y, columns=['count'])['count'].value_counts()
    df = pd.DataFrame(df, index=df.index, columns=['count'])
    df['p'] = df['count']/sum(df['count'])
    display(df)
    # print(get_cls_dist_string(self.cls, self.c_szs, self.c_dist))


# def calc_norm(self):
#     xs, ys = [], []
#     for i in range(len(self)):
#         x, y = self[i]
#         xs.append(x)
#         ys.append(y)
#     x, y = torch.stack(xs, 0), torch.stack(ys, 0)
#     mean = [x[:,i,:,:].mean().item() for i in range(3)]
#     std = [x[:,i,:,:].std().item() for i in range(3)]
#     return mean, std



# # L()

# def is_iter(o):
#     "Test whether `o` can be used in a `for` loop"
#     #Rank 0 tensors in PyTorch are not really iterable
#     return isinstance(o, (Iterable,Generator)) and getattr(o,'ndim',1)


# def mask2idxs(mask):
#     "Convert bool mask or index list to index `L`"
#     if isinstance(mask,slice): return mask
#     mask = list(mask)
#     if len(mask)==0: return []
#     it = mask[0]
#     if hasattr(it,'item'): it = it.item()
#     if isinstance(it,(bool,NoneType,np.bool_)): return [i for i,m in enumerate(mask) if m]
#     #if isinstance(it,(bool,np.bool_)): return [i for i,m in enumerate(mask) if m]
#     return [int(i) for i in mask]


# def _is_array(x): return hasattr(x,'__array__') or hasattr(x,'iloc')


# def is_indexer(idx):
#     "Test whether `idx` will index a single item in a list"
#     return isinstance(idx,int) or not getattr(idx,'ndim',1)


# def _listify(o):
#     if o is None: return []
#     if isinstance(o, list): return o
#     if isinstance(o, str) or _is_array(o): return [o]
#     if is_iter(o): return list(o)
#     return [o]


# def coll_repr(c, max_n=10):
#     "String repr of up to `max_n` items of (possibly lazy) collection `c`"
#     return f'(#{len(c)}) [' + ','.join(itertools.islice(map(str,c), max_n)) + (
#         '...' if len(c)>10 else '') + ']'


# class CollBase:
#     "Base class for composing a list of `items`"
#     def __init__(self, items): self.items = items
#     def __len__(self): return len(self.items)
#     def __getitem__(self, k): return self.items[k]
#     def __setitem__(self, k, v): self.items[list(k) if isinstance(k,CollBase) else k] = v
#     def __delitem__(self, i): del(self.items[i])
#     def __repr__(self): return self.items.__repr__()
#     def __iter__(self): return self.items.__iter__()


# class L(CollBase):
#     "Behaves like a list of `items` but can also index with list of indices or masks"
#     _default='items'
#     def __init__(self, items=None, *rest, use_list=False, match=None):
#         if rest: items = (items,)+rest
#         if items is None: items = []
#         if (use_list is not None) or not _is_array(items):
#             items = list(items) if use_list else _listify(items)
#         if match is not None:
#             if is_coll(match): match = len(match)
#             if len(items)==1: items = items*match
#             else: assert len(items)==match, 'Match length mismatch'
#         super().__init__(items)

#     @property
#     def _xtra(self): return None
#     def _new(self, items, *args, **kwargs): return type(self)(items, *args, use_list=None, **kwargs)
#     def __getitem__(self, idx): return self._get(idx) if is_indexer(idx) else L(self._get(idx), use_list=None)
#     def copy(self): return self._new(self.items.copy())

#     def _get(self, i):
#         if is_indexer(i) or isinstance(i,slice): return getattr(self.items,'iloc',self.items)[i]
#         i = mask2idxs(i)
#         return (self.items.iloc[list(i)] if hasattr(self.items,'iloc')
#                 else self.items.__array__()[(i,)] if hasattr(self.items,'__array__')
#                 else [self.items[i_] for i_ in i])

#     def __setitem__(self, idx, o):
#         "Set `idx` (can be list of indices, or mask, or int) items to `o` (which is broadcast if not iterable)"
#         idx = idx if isinstance(idx,L) else _listify(idx)
#         if not is_iter(o): o = [o]*len(idx)
#         for i,o_ in zip(idx,o): self.items[i] = o_

#     def __iter__(self): return iter(self.items.itertuples() if hasattr(self.items,'iloc') else self.items)
#     def __contains__(self,b): return b in self.items
#     def __invert__(self): return self._new(not i for i in self)
#     def __eq__(self,b): return False if isinstance(b, (str,dict,set)) else all_equal(b,self)
#     def __repr__(self): return repr(self.items) if _is_array(self.items) else coll_repr(self)
#     def __mul__ (a,b): return a._new(a.items*b)
#     def __add__ (a,b): return a._new(a.items+_listify(b))
#     def __radd__(a,b): return a._new(b)+a
#     def __addi__(a,b):
#         a.items += list(b)
#         return a

#     def sorted(self, key=None, reverse=False):
#         if isinstance(key,str):   k=lambda o:getattr(o,key,0)
#         elif isinstance(key,int): k=itemgetter(key)
#         else: k=key
#         return self._new(sorted(self.items, key=k, reverse=reverse))

#     @classmethod
#     def split(cls, s, sep=None, maxsplit=-1): return cls(s.split(sep,maxsplit))

#     @classmethod
#     def range(cls, a, b=None, step=None):
#         if is_coll(a): a = len(a)
#         return cls(range(a,b,step) if step is not None else range(a,b) if b is not None else range(a))

#     def map(self, f, *args, **kwargs):
#         g = (bind(f,*args,**kwargs) if callable(f)
#              else f.format if isinstance(f,str)
#              else f.__getitem__)
#         return self._new(map(g, self))

#     def filter(self, f, negate=False, **kwargs):
#         if kwargs: f = partial(f,**kwargs)
#         if negate: f = negate_func(f)
#         return self._new(filter(f, self))

#     def unique(self): return L(dict.fromkeys(self).keys())
#     def enumerate(self): return L(enumerate(self))
#     def val2idx(self): return {v:k for k,v in self.enumerate()}
#     def itemgot(self, *idxs):
#         x = self
#         for idx in idxs: x = x.map(itemgetter(idx))
#         return x
    
#     def attrgot(self, k, default=None): return self.map(lambda o:getattr(o,k,default))
#     def cycle(self): return cycle(self)
#     def map_dict(self, f=noop, *args, **kwargs): return {k:f(k, *args,**kwargs) for k in self}
#     def starmap(self, f, *args, **kwargs): return self._new(itertools.starmap(partial(f,*args,**kwargs), self))
#     def zip(self, cycled=False): return self._new((zip_cycle if cycled else zip)(*self))
#     def zipwith(self, *rest, cycled=False): return self._new([self, *rest]).zip(cycled=cycled)
#     def map_zip(self, f, *args, cycled=False, **kwargs): return self.zip(cycled=cycled).starmap(f, *args, **kwargs)
#     def map_zipwith(self, f, *rest, cycled=False, **kwargs): return self.zipwith(*rest, cycled=cycled).starmap(f, **kwargs)
#     def concat(self): return self._new(itertools.chain.from_iterable(self.map(L)))
#     def shuffle(self):
#         it = copy(self.items)
#         random.shuffle(it)
#         return self._new(it)
    
#     def append(self,o): return self.items.append(o)
#     def remove(self,o): return self.items.remove(o)
#     def count (self,o): return self.items.count(o)
#     def reverse(self ): return self.items.reverse()
#     def pop(self,o=-1): return self.items.pop(o)
#     def clear(self   ): return self.items.clear()
#     def index(self, value, start=0, stop=sys.maxsize): return self.items.index(value, start=start, stop=stop)
#     def sort(self, key=None, reverse=False): return self.items.sort(key=key, reverse=reverse)



import numpy as np
import pickle
from dlogicutils.df import df_add_datepart


def scaller_norm(s, _mean, _std): return (s-_mean)/_std if _std>0 else s

def scaller_minmax(s, _min, _max): return ((s-_min)/(_max-_min)-0.5)*2.0 if (_max-_min)>0 else s

def scaller_rank(s):
    raise NotImplementedError


class DataFrameProcessor():
    """
    Procecesses Dataframe:
        Transform dates into dateparts
        Split cat and cont vars
        Transform low cardinality cat vars into dummies
        Include NA cols for cont vars
        Calculates stats and normalizes data with a scaller
        Converts cat vars into int32, cont vars into float32, depvar into float32

        Strores processing state for future application on different dataframes
        Saves and loads state into pickle
    """
    
    def __init__(self, c_cat, c_cont, c_dep, c_date, c_key, c_skip, c_onehot, c2, ncats, nconts, cat_szs, emb_szs,
                    cat_dicts, means, stds, medians, mins, maxs, maxhot):
        self.c_cat, self.c_cont, self.c_dep, self.c_date, self.c_key, self.c_skip = c_cat, c_cont, c_dep, c_date, c_key, c_skip
        self.maxhot, self.c_onehot, self.ncats, self.nconts = maxhot, c_onehot, ncats, nconts
        self.cat_szs, self.emb_szs, self.cat_dicts = cat_szs, emb_szs, cat_dicts,
        self.means, self.stds, self.medians, self.mins, self.maxs = means, stds, medians, mins, maxs
        self.c = self.c_skip + self.c_cat + self.c_onehot + self.c_cont
        if c_key is not None: self.c = [c_key] + self.c
        self.c2 = c2
    
    @classmethod
    def from_df(cls, df, c_cat, c_cont, c_dep, c_date=[], c_key=None, c_skip=[], max_embs=50, maxhot=16, embs={}):
        print(f'DataFrameProcessor: building meta')
        df = df.copy()
        # Change args lists access from reference to value
        c_cat, c_cont, c_date = c_cat[:], c_cont[:], c_date[:]
        # Transform dates in dateparts
        for c in c_date:
            c_num = [f'{c}_Dayofyear', f'{c}_Week', f'{c}_Day']
            c_new = df_add_datepart(df, c, drop=False)
            c_cat += [e for e in c_new if e not in c_num]
            c_cont += c_num
        # Calc cardinality of all cats
        cat_szs = {c: len(df[c].unique()) for c in c_cat}
        # For cats with cardinality <= maxhot, transforms into onehot
        c_onehot = [c for c in c_cat if cat_szs[c] <= maxhot]
        # Exclude onehots from cats
        c_cat = [c for c in c_cat if c not in c_onehot]
        cat_szs = {c: cat_szs[c] for c in c_cat}
        # Define embeddings sizes
        emb_szs = [(cat_szs[c]+1, embs[c] if c in embs else min((cat_szs[c]+1)//2, max_embs)) for c in cat_szs]
        # fast.ai heuristic: min(round(1.6*cat_szs[c]**0.56), max_embs)
        # Build one dict (key->index) for all cats (missing values correspond to zero)
        cat_dicts = {c: {} for c in c_cat+c_onehot}
        for c in c_cat+c_onehot:
            cat_dicts[c] = {e: i+1 for i,e in enumerate(list(np.sort(df[df[c].isna()==False][c].astype(str).unique())))}
        # Calculate number of cats and number of conts including onehots
        ncats = len(c_cat)
        nconts = len(c_cont*2)
        for c in c_onehot:
            nconts += len(df[df[c].isna()==False][c].unique())+1
        # Calculate statistics for all conts
        means = {c: df[c].mean() for c in c_cont}
        stds = {c: df[c].std() for c in c_cont}
        medians = {c: df[c].median() for c in c_cont}
        mins = {c: df[c].min() for c in c_cont}
        maxs = {c: df[c].max() for c in c_cont}
        print(f'Done\n')
        return cls(c_cat, c_cont, c_dep, c_date, c_key, c_skip, c_onehot, [], ncats, nconts, cat_szs, emb_szs,
                    cat_dicts, means, stds, medians, mins, maxs, maxhot)
        
    @classmethod
    def from_meta_file(cls, fname):
        meta = pickle.load(open(fname, 'rb'))
        return cls(meta['c_cat'], meta['c_cont'], meta['c_dep'], meta['c_date'],  meta['c_key'], meta['c_skip'],
                    meta['c_onehot'], meta['c2'], meta['ncats'], meta['nconts'], meta['cat_szs'], meta['emb_szs'], 
                    meta['cat_dicts'], meta['means'], meta['stds'], meta['medians'], meta['mins'], meta['maxs'],
                    meta['maxhot'])

    def __call__(self, df):
        print(f'DataFrameProcessor: processing shape={df.shape}')
        df = df.copy()
        df = df.replace([np.inf, -np.inf], np.nan)
        for c in self.c_date: _ = df_add_datepart(df, c, drop=True)
        # df = df[self.c]
        for c in self.c:
            un = len(df[c].unique())
            if un<2: print(f'Warning: column {c} with {un} unique values')
        for c in self.c_cat+self.c_onehot:
            df[c] = df[c].astype(str)
            df[c] = [self.cat_dicts[c][e] if e in self.cat_dicts[c] else 0 for e in df[c]]
        c_dummies = []
        for c in self.c_onehot:
            #if len(self.cat_dicts[c]) <= 2:
            #    c_dummies.append(c)
            #    continue
            for e in self.cat_dicts[c]:
                df[c+'_'+e] = (df[c]==self.cat_dicts[c][e])
                c_dummies.append(c+'_'+e)
            df[c+'_na'] = df[c].isna()
            c_dummies.append(c+'_na')
        c_na = []
        for c in self.c_cont:
            # if sum(df[c].isna())==0: continue
            c_na.append(c+'_na')
            df[c+'_na'] = df[c].isna()
            df[c] = df[c].fillna(self.medians[c])
            df[c] = df[c].replace([np.inf, -np.inf], 0.0)
            df[c] = scaller_norm(df[c], self.means[c], self.stds[c])
        c_cont2 = self.c_cont + c_na + c_dummies
        for c in self.c_cat: df[c] = df[c].astype('int32')
        for c in c_cont2: df[c] = df[c].astype('float32')
        self.c2 = self.c_cat + c_cont2
        df = df[self.c2]
        print(f'Done: ncats={len(self.c_cat)}, nconts={len(c_cont2)} ' +\
              f'(numeric={len(c_cont2)-len(c_dummies)-len(c_na)}, na={len(c_na)}, dummies={len(c_dummies)})\n')
        return df
    
    def save_meta(self, fname):
        df_proc_meta = {
            'c_cat': self.c_cat, 'c_cont': self.c_cont, 'c_dep': self.c_dep,
            'c_date': self.c_date, 'c_key': self.c_key, 'c_skip': self.c_skip, 'c_onehot': self.c_onehot, 'c2': self.c2,
            'ncats': self.ncats, 'nconts': self.nconts, 'cat_szs': self.cat_szs, 'emb_szs': self.emb_szs,
            'cat_dicts': self.cat_dicts, 'means': self.means, 'stds': self.stds, 'medians': self.medians,
            'mins': self.mins, 'maxs': self.maxs, 'maxhot': self.maxhot,  
        }
        pickle.dump(df_proc_meta, open(fname, 'wb'))

    def get_emb_szs(self):
        return sorted(list(zip(self.cat_szs, self.emb_szs)), key = lambda e: -e[1][0])


def df_sample(df, n=1000): return df.sample(n=n).copy()
def df_shufle(df): return df.sample(frac=1)


def df_split(df, p1=0.8, p3=None):
    n = len(df)
    perm = np.random.permutation(n)
    cut1 = int(n*p1)
    if p3 is None:
        return df.iloc[perm[:cut1]].copy(), df.iloc[perm[cut1:]].copy()
    cut2 = int(n*(1.-p3))
    return df.iloc[perm[:cut1]].copy(), df.iloc[perm[cut1:cut2]].copy(),  df.iloc[perm[cut2:]].copy()


def df_split_by_date(df, date1, date2=None, date3=None, c_date='date'):
    if date2 is None:
        return df[df[c_date]<date1].copy(), df[df[c_date]>=date1].copy()
    if date3 is None:
        return df[df[c_date]<date1].copy(), df[(df[c_date]>=date1)&(df[c_date]<date2)].copy(), df[df[c_date]>=date2].copy()
    return df[df[c_date]<date1].copy(), df[(df[c_date]>=date1)&(df[c_date]<date2)].copy(), df[(df[c_date]>=date2)&(df[c_date]<date3)].copy()



import random
import feather
import numpy as np
import torch
import torchvision.transforms as transforms
from PIL import Image
from IPython.display import display
from dlogicutils.core import *
from dlogicutils.fs import get_dir_files, get_dir_dirs
from nn2.etc import *
from nn2.tabular import DataFrameProcessor

# if regr:
#     yrange = [df[c_dep].astype('float32').min(), df[c_dep].astype('float32').max()]


################################################################################################################################
#  DataLoader
#
# TODO: num_workers, pin_memory
################################################################################################################################

def get_balanced_idxs(c_idxs, p=0.):
    p = min(max(p,0.),1.)
    szs = np.array([len(e) for e in c_idxs])
    n = szs.sum()
    bp0 = np.array([len(c)/n for c in c_idxs])
    bp1 = np.array([1/len(c_idxs)]*len(c_idxs))
    bp = bp0+p*(bp1-bp0)
    m = bp/bp0
    szs2 = np.round((szs*m/m.min())).astype(int)
    z = np.array([])
    for i in range(len(c_idxs)):
        for _ in range(szs2[i]//szs[i]): z = np.concatenate((z,np.array(c_idxs[i])))
        z = np.concatenate((z,np.array(c_idxs[i])[:szs2[i]%szs[i]]))
    z = np.random.permutation(z)
    return z.astype(int)


def collate(b):
    x,y = zip(*b)
    return torch.stack(x),torch.stack(y)

class SimpleSampler():
    def __init__(self, ds, bs):
        self.n, self.bs = len(ds), bs
    def set_balance(self, p): pass
    def __iter__(self):
        self.idxs = list(range(0,self.n))
        for i in range(0, self.n, self.bs): yield self.idxs[i:i+self.bs]

class RandomSampler():
    def __init__(self, ds, bs, c_idxs): self.n, self.bs, self.c_idxs, self.p = len(ds), bs, c_idxs, None
    def set_balance(self, p): self.p = p
    def __iter__(self):
        if self.p is None:
            self.idxs = list(np.random.permutation(np.arange(self.n)))
        else:
            self.idxs = list(get_balanced_idxs(self.c_idxs, p=self.p))
        for i in range(0, self.n, self.bs): yield self.idxs[i:i+self.bs]


class DataLoader():
    def __init__(self, ds, bs=64, shuffle=False, collate_fn=collate):
        self.ds, self.bs, self.collate_fn, self.n = ds, bs, collate_fn, len(ds)
        self.sampler = SimpleSampler(ds, bs) if not shuffle else RandomSampler(ds, bs, self.ds.c_idxs)
    def set_balance(self, p): self.sampler.set_balance(p)
    def __len__(self): return math.ceil(len(self.ds)/self.bs)
    def __iter__(self):
        for b in self.sampler: yield self.collate_fn([self.ds[i] for i in b])


################################################################################################################################
#  Dataset, Data
################################################################################################################################

class Dataset():
    def __init__(self, x, y, clsf=False, cls_to_idx=None, regr=False, trfs=[]):
        self.x, self.y, self.clsf, self.regr, self.trfs = x, y, clsf, regr, trfs or []
        self.bps = None
        self.cls, self.c, self.cls_to_idx, self.c_idxs, self.c_szs, self.c_dist = None, None, None, None, None, None
        if clsf:
            assert type(y)==list
            self.labels = [str(e) for e in y]
            if cls_to_idx is None:
                self.cls = sorted(list(set(self.labels)))
                self.cls_to_idx = {self.cls[i]: i for i in range(len(self.cls))}
            else:
                self.cls = [e for e in cls_to_idx]
                self.cls_to_idx = cls_to_idx
                for label in self.labels:
                    if label not in self.cls:
                        print(f'Label {label} not found in cls_to_idx')
                        raise Exception()
            self.c = len(self.cls)
            self.y = torch.tensor([self.cls_to_idx[label] for label in self.labels])
            self.c_idxs, self.c_szs, self.c_dist = get_cls_dist(self.labels, self.cls_to_idx)
            print(f'\n{self.__class__.__name__} for Classification: size: {len(self.y)}, cls: {self.c}')
            show_dist(self.labels)
    def __len__(self): return len(self.x)
    def __getitem__(self, i):
        x,y = self.load(self.x[i], self.y[i])
        for trf in self.trfs: x = trf(x)
        return x,y
    def load(self, x, y): return x,y
    def show(self, idxs, labels=None):
        for idx in listify(idxs):
            x,y = self[idx]
            print(x.shape, y.shape)
    def show_trfs(self, n=8): self.show([random.randint(0,len(self))]*n)
    def encode(self, x, y): pass
    def decode(self,x,y): return x.numpy(),y.item()


class Data:
    def __init__(self, ds1, ds2, clsf=False, regr=False, bs=64):
        self.ds1, self.ds2 =  ds1, ds2
        self.clsf, self.regr, self.bs = clsf, regr, bs
        # w, pin = 0, torch.cuda.is_available()
        self.dl1 = DataLoader(ds1, bs=bs, shuffle=True)
        self.dl2 = DataLoader(ds2, bs=bs, shuffle=False) if ds2 is not None else None

    def show_batch(self, n=8): self.ds1.show(next(iter(self.dl1.sampler))[:n])
    def show_batch_2(self, n=8): self.ds2.show(next(iter(self.dl2.sampler))[:n])
    def show_trfs(self, n=8): self.ds1.show_trfs(n=n)

    @property
    def c(self): return self.ds1.c if self.clsf else None
    @property
    def cls(self): return self.ds1.cls if self.clsf else None
    @property
    def cls_to_idx(self): return self.ds1.cls_to_idx if self.clsf else None


################################################################################################################################
#  ImageDataset, ImageDatasetMem, ImageData
################################################################################################################################

class ImageDataset(Dataset):    
    def load(self,x,y): return Image.open(x),y
    def decode(self, x, y): return transforms.ToPILImage()(x),y
    def show(self, idxs, labels=None):
        idxs, labels = listify(idxs), listify(labels)
        x,y = zip(*[self.decode(*self[idx]) for idx in idxs])
        x = np.stack([np.array(x) for x in x])
        if self.clsf: y = [self.cls[e] for e in y]
        imgrid(x, labels or y)


class ImageDatasetMem(ImageDataset):
    def load(self,x,y): return x,y
    def decode(self,x,y): return x.numpy().transpose((1,2,0)),y


class ImageData(Data):
    def __init__(self, x1, y1, x2, y2, clsf=False, regr=False, trfs1=[], trfs2=[], bs=64, ds_class=ImageDataset):
        ds1 = ds_class(x1, y1, clsf=clsf, regr=regr, trfs=trfs1)
        ds2 = ds_class(x2, y2, clsf=clsf, regr=regr, trfs=trfs2, cls_to_idx=ds1.cls_to_idx)
        super(ImageData, self).__init__(ds1, ds2, clsf=clsf, regr=regr, bs=bs)
        
    @classmethod
    def from_folder(cls, path1, path2, types=None, clsf=False, regr=False, trfs1=[], trfs2=[], bs=64):
        x, y = [[],[]], [[],[]]
        cls_names = [d.stem for d in get_dir_dirs(path1)]
        for i,path in enumerate([path1, path2]):
            for c in cls_names:
                for f in get_dir_files(path/f'{c}', types):
                    x[i].append(f)
                    y[i].append(c)
        return cls(x[0], y[0], x[1], y[1], clsf=clsf, regr=regr, trfs1=trfs1, trfs2=trfs2, bs=bs)

    @classmethod
    def from_numpy(cls, x1, y1, x2, y2, clsf=False, regr=False, trfs1=[], trfs2=[], bs=64):
        x1, y1, x2, y2 = T(x1).float(), y1, T(x2).float(), y2
        return cls(x1, y1, x2, y2, clsf=clsf, regr=regr, trfs1=trfs1, trfs2=trfs2, bs=bs, ds_class=ImageDatasetMem)

################################################################################################################################
#  SignalDataset, SignalData
################################################################################################################################

class SignalDataset(Dataset):    
    def show(self, idxs, labels=None):
        idxs, labels = listify(idxs), listify(labels)
        for i,idx in enumerate(idxs):
            x,y = self.decode(*self[idx])
            if self.clsf: y = self.cls[y]
            label = labels[i] if len(labels) else y
            sigshow(x, label or y)

class SignalData(Data):
    def __init__(self, x1, y1, x2, y2, clsf=False, regr=False, trfs1=[], trfs2=[], bs=64):
        ds1 = SignalDataset(x1, y1, clsf=clsf, regr=regr, trfs=trfs1)
        ds2 = SignalDataset(x2, y2, clsf=clsf, regr=regr, trfs=trfs2)
        super(SignalData, self).__init__(ds1, ds2, clsf=clsf, regr=regr, bs=bs)
        
    @classmethod
    def from_numpy(cls, x1, y1, x2, y2, clsf=False, regr=False, trfs1=[], trfs2=[], bs=64):
        x1, x2 = torch.from_numpy(x1).float(), torch.from_numpy(x2).float()
        if clsf: y1, y2 = list(y1), list(y2)
        else: y1, y2 = torch.from_numpy(y1).float(), torch.from_numpy(y2).float()
        return cls(x1, y1, x2, y2, clsf=clsf, regr=regr, trfs1=trfs1, trfs2=trfs2, bs=bs)


################################################################################################################################
#   TabularDataset, TabularData
################################################################################################################################

class TabularDataset(Dataset):
    def __init__(self, df, dfp, proc, clsf=False, cls_to_idx=None, regr=False):
        self.df, self.dfp = df, dfp
        x  = dfp[proc.c2].values
        x = torch.from_numpy(x).float()
        y = [str(e) for e in df[proc.c_dep].values]
        super(TabularDataset, self).__init__(x, y, clsf=clsf, cls_to_idx=cls_to_idx, regr=regr)
        if len(proc.get_emb_szs()): print(f'Embeddings: {proc.get_emb_szs()}')

    def show(self, idxs, labels=None, cols=None):
        df = self.df.iloc[listify(idxs)][cols or self.df.columns]
        df['results'] = labels
        display(df)


class TabularData(Data):
    def __init__(self, df1, df1p, df2, df2p, proc, clsf=False, regr=False, bs=128):
        self.df1, self.df1p, self.df2, self.df2p, self.proc = df1, df1p, df2, df2p, proc
        ds1 = TabularDataset(df1, df1p, proc, clsf=clsf, regr=regr)
        ds2 = TabularDataset(df2, df2p, proc, clsf=clsf, regr=regr,  cls_to_idx=ds1.cls_to_idx) if df2p is not None else None
        super(TabularData, self).__init__(ds1, ds2, clsf=clsf, regr=regr, bs=bs)

    @classmethod
    def from_dfs(cls, df1, df2, c_cat, c_cont, c_dep, c_date=[], c_key=None, c_skip=[], maxhot=16, max_embs=4, embs={}, path='.', 
                    clsf=False, regr=False, name='tabular', bs=256):
        proc = DataFrameProcessor.from_df(df1, c_cat, c_cont, c_dep, c_date=c_date, c_key=c_key, c_skip=c_skip, maxhot=maxhot,
                                            max_embs=max_embs, embs=embs)
        df1p = proc(df1)
        proc.save_meta(f'{path}/{name}_proc.pickle')
        df1.reset_index(drop=True).to_feather(f'{path}/{name}_df1.feather')
        df1p.reset_index(drop=True).to_feather(f'{path}/{name}_df1p.feather')
        df2p = None
        if df2 is not None:
            df2p = proc(df2)
            df2.reset_index(drop=True).to_feather(f'{path}/{name}_df2.feather')
            df2p.reset_index(drop=True).to_feather(f'{path}/{name}_df2p.feather')
        return cls(df1, df1p, df2, df2p, proc, clsf=clsf, regr=regr, bs=bs)

    @classmethod
    def from_feather(cls, path='.', name='tabular', clsf=False, regr=False, bs=128, valid=True):
        proc = DataFrameProcessor.from_meta_file(f'{path}/{name}_proc.pickle')
        df1 = feather.read_dataframe(f'{path}/{name}_df1.feather')
        df1p = feather.read_dataframe(f'{path}/{name}_df1p.feather')
        if not valid: return cls(df1, df1p, None, None, proc, clsf=clsf, regr=regr, bs=bs)
        df2, df2p = None, None
        if Path(f'{path}/{name}_df2.feather').is_file():
            df2 = feather.read_dataframe(f'{path}/{name}_df2.feather')
            df2p = feather.read_dataframe(f'{path}/{name}_df2p.feather')
        return cls(df1, df1p, df2, df2p, proc, clsf=clsf, regr=regr, bs=bs)

    def encode(self, df): return torch.from_numpy(self.proc(df)[self.proc.c2].values).float()
    def decode(self): pass

    @property
    def ncats(self): return self.proc.ncats
    @property
    def nconts(self): return self.proc.nconts
    @property
    def maxhot(self): return self.proc.maxhot
    @property
    def cat_szs(self): return self.proc.cat_szs
    @property
    def emb_szs(self): return self.proc.emb_szs



import numpy as np
from sklearn.metrics import confusion_matrix
import matplotlib.pyplot as plt
import matplotlib as mpl
from nn2.data import TabularDataset
from nn2.etc import *
from IPython.display import display
import sklearn
# https://en.wikipedia.org/wiki/Sensitivity_and_specificity

epsilon = 1e-9
scols = ['p','n','tp','fp','tn','fn','tpr','fpr','tnr','fnr','accuracy','error_rate','precision','recall',
         'sensitivity','specificity']

def df_stats_by_class(cm):
    rows = []
    for i in range(cm.shape[0]):
        precision = cm[i,i]/cm[i,:].sum()
        recall = cm[i,i]/(cm[:,i].sum()+epsilon)
        f1 = 2*(precision*recall)/(precision+recall+epsilon)
        support = cm[:,i].sum()
        rows.append((precision, recall, f1, support))
    return pd.DataFrame(rows, columns=['precision', 'recall', 'f1', 'support'])

def get_stats(preds, actuals, th=.5):
    p = np.sum(actuals==1)
    n = np.sum(actuals==0)
    tp = np.sum((preds==1)*(actuals==1)*1)
    fp = np.sum((preds==1)*(actuals==0)*1)
    tn = np.sum((preds==0)*(actuals==0)*1)
    fn = np.sum((preds==0)*(actuals==1)*1)
    tpr = tp/(p+epsilon)
    fpr = fp/(n+epsilon)
    tnr = tn/(n+epsilon)
    fnr = fn/(p+epsilon)
    accuracy = np.sum(preds==actuals)/(len(actuals))
    error_rate = 1.-accuracy
    precision = tp/(tp+fp+epsilon)
    recall = tpr
    sensitivity = recall
    specificity = tnr
    return [p,n,tp,fp,tn,fn,tpr,fpr,tnr,fnr,accuracy,error_rate,precision,recall,sensitivity,specificity], scols

def df_stats_binary(preds, actuals, th=.5):
    stats, stats_cols = get_stats(preds, actuals, th=.5)
    stats = [len(actuals)]+stats
    cols = ['total']+stats_cols
    return pd.DataFrame([stats], columns=cols)

def df_stats_cum(probs, actuals, th=.5, thresholds=None):
    mask = np.argsort(-probs)
    actuals, probs = actuals[mask], probs[mask]
    preds = (probs>=th)*1
    ptotal = np.sum(actuals==1)
    nn = np.arange(len(actuals))+1
    nnp = nn/len(actuals)
    p = np.cumsum(actuals==1)
    n = np.cumsum(actuals==0)
    tp = np.cumsum((preds==1)*(actuals==1)*1)
    fp = np.cumsum((preds==1)*(actuals==0)*1)
    tn = np.cumsum((preds==0)*(actuals==0)*1)
    fn = np.cumsum((preds==0)*(actuals==1)*1)
    tpr = tp / (p+epsilon)
    fpr = fp / (n+epsilon)
    tnr = tn / (n+epsilon)
    fnr = fn / (p+epsilon)
    accuracy = np.cumsum(preds==actuals)/(nn)
    error_rate = 1.-accuracy
    precision = tp/(tp+fp+epsilon)
    recall = tpr
    sensitivity = recall
    specificity = tnr
    pcaptured = tp/ptotal
    z = zip(probs,preds,actuals,nn,nnp,p,n,tp,fp,tn,fn,tpr,fpr,tnr,fnr,accuracy,error_rate,precision,recall,sensitivity,specificity,pcaptured)
    cols = ['prob','pred','actual','size','size%']+scols+['pcaptured']
    df = pd.DataFrame(z, columns=cols)
    if thresholds:
        ths = [t/thresholds if t>0 else -.00001 for t in range(0,thresholds)[::-1]]
        idxs = [max(np.where(probs>t)[0]) for t in ths]
        df = df.iloc[idxs].copy()
        df.insert(0, 'threshold', ths)
        df.drop(columns=['prob','pred','actual'], inplace=True)
    return df

def plot_cmap(data):
    fig, ax = plt.subplots(figsize=(8, 8))
    cmap=sns.diverging_palette(0, 160, as_cmap=True)
    cmap=mpl.cm.Blues
    cmap = sns.diverging_palette(0, 180, sep=40, s=100, as_cmap=True)
    sns.heatmap(data, vmin=-1, vmax=1, cmap=cmap, square=True, ax=ax)
    plt.show()
    plt.pause(.001)

def plot_confusion(c, actuals, preds):
    cm = np.zeros((c, c), dtype=int)
    cm_p = np.zeros((c, c), dtype=np.float)
    for i in range(preds.shape[0]): cm[int(actuals[i])][int(preds[i])] += 1
    for i in range(cm.shape[0]): cm_p[i,:] = cm[i,:] / cm[i,:].sum()
    cm_p = 100.0*cm_p
    fig = plt.figure(figsize=(20, 5))
    ax1, ax2 = fig.add_subplot(121), fig.add_subplot(122)
    sns.set(font_scale=1.0)
    sns.heatmap(cm, cmap=mpl.cm.Blues, annot=True, fmt='d', ax=ax1)
    sns.heatmap(cm_p, cmap=mpl.cm.Blues, annot=True, fmt='.2f', ax=ax2)
    ax1.set_title(f'Total: {preds.shape[0]}')
    ax2.set_title(f'Total: 100.00%')
    plt.show()
    plt.pause(.001)
    return cm

def calc_prc(y, y2):
    mask = np.argsort(-y2)
    y, y2 = y[mask], y2[mask]
    recall = np.array([0]+[((y2>=t)[y==1]==y[y==1]).sum()/(y==1).sum() for t in y2]+[1])
    precision = np.array([1]+[(y[y2>=t]==1).sum()/(y2>=t).sum() for t in y2]+[0])
    auc_pr = np.trapz(precision, recall)
    return auc_pr, precision, recall

def calc_roc(y, y2):
    mask = np.argsort(-y2)
    y, y2 = y[mask], y2[mask]
    fpr = np.array([0]+[((y2>=t)[y==0]!=y[y==0]).sum()/(y==0).sum() for t in y2])
    tpr = np.array([0]+[((y2>=t)[y==1]==y[y==1]).sum()/(y==1).sum() for t in y2])
    auc = np.trapz(tpr, fpr)
    return auc, fpr, tpr

def plot_roc(y, y2):
    fpr,tpr,_ = sklearn.metrics.roc_curve(y, y2)
    auc = auc = np.trapz(tpr, fpr)
    precision,recall,_ = sklearn.metrics.precision_recall_curve(y, y2)
    auc_pr = np.trapz(precision, recall)
    # auc, fpr, tpr = calc_roc(y, y2)
    # auc_pr, precision, recall = calc_prc(y, y2)
    _, axs = plt.subplots(1,2, figsize=(14,5))
    axs[0].plot(fpr, tpr, label=f'AUC = {auc:.3f})')
    axs[0].plot([0., 1.], [0., 1.], 'k--')
    axs[0].set_title('ROC')
    axs[0].set_xlabel('FPR')
    axs[0].set_ylabel('TPR = Recall')
    axs[0].legend(loc='best')
    axs[1].plot(recall, precision,  label=f'AUC_PR = {auc_pr:.3f})')
    axs[1].plot([0., 1.], [1., 0.], 'k--')
    axs[1].set_title('ROC - Precision/Recall')
    axs[1].set_xlabel('Recall')
    axs[1].set_ylabel('Precision')
    axs[1].legend(loc='best')
    plt.show()
    plt.pause(.001)

def get_corr(df, cols, c=None, plot=False, max_vars=None):
    dfc = df[cols].corr()
    if c is not None:
        dfc=dfc[c].copy()
        n = max_vars or len(dfc)
        dfc = pd.DataFrame(dfc.reindex(dfc.abs().sort_values(ascending=False).index)[:n])
        dfc = dfc.drop(index=[c])
    display(dfc)
    if plot: plot_cmap(dfc)

def get_probs(y2): return np.exp(np.max(y2, axis=1))
def get_probs_bin(y2): return np.exp(y2[:,1])
def get_preds(probs, t=.5): return (probs>t)*1

def get_predictions_clsf(model, x, device='cpu'):
    model.eval()
    with torch.no_grad():
        x = x.to(device)
        y2 = model(x)
    return y2.detach().cpu().numpy()
# x = torch.cat([x for x,y in learner.data.dl2])

def get_predictions_clsf_tabular(data, model, df, device='cpu'):
    dfp = data.proc(df.copy())
    x = dfp[data.proc.c2].values
    x = torch.from_numpy(x).float()
    y2 = get_predictions_clsf(model, x, device)
    prob = get_probs(y2) if data.c>2 else get_probs_bin(y2)
    pred = get_preds(prob)
    df['prob'] = prob
    df['pred'] = pred
    df['pred_label'] = [data.cls[e] for e in pred]
    return df


# ???
def get_probs_for_actuals(y2, actuals):
    probs = np.exp(y2)
    mask = np.zeros((probs.shape[0], probs.shape[1]))
    mask[np.arange(actuals.shape[0]),actuals] = 1
    return probs[mask.astype(bool)]


# ???
def df_predict_clsf(learner, df, have_actuals=False):
    df = df.copy()
    if 'prob' in df.columns: df=df.drop(columns='prob')
    data = learner.data
    y2 = learner.predict(df)
    y2 = y2.numpy()
    prob = get_probs(y2) if data.c>2 else get_probs_bin(y2)
    pred = get_preds(y2)
    pred_label = [data.cls[e] for e in pred]
    df['prob'] = prob
    df['pred'] = pred
    df['pred_label'] = pred_label
    if have_actuals:
        actuals = df[data.proc.c_dep] 
        df['actual'] = actuals.values
        df['actual_label'] = [data.cls[e] for e in actuals.values]
        df['error'] = (df['pred']!=df['actual'])*1
    df = df.sort_values(by=['prob'], ascending=[False])
    return df



class Results_Clsf():
    def __init__(self, data, probs, actuals):
        self.data, self.probs, self.actuals = data, probs, actuals
        self.preds = get_preds(probs)
#         self.probs_for_actuals = get_probs_for_actuals(y2, self.actuals)
#         self.probs_sorted = np.argsort(self.probs_for_actuals)[::-1]
    def show_all(self, thresholds=10):
        self.confusion()
        self.roc()
        display(self.df_stats(thresholds=thresholds))
    def confusion(self, t=None):
        show_dist(self.actuals)
        display(df_stats_binary(self.preds, self.actuals))
        if t is None:
            preds = self.preds
        else:
            if self.data.c!=2: raise ValueError('More than two classes')
            preds = (self.probs>t)*1
        cm = plot_confusion(self.data.c, self.actuals, preds)
        display(df_stats_by_class(cm))
    def df_stats(self, thresholds=10):
        if self.data.c!=2: raise ValueError('More than two classes')
        return df_stats_cum(self.probs, self.actuals, thresholds=thresholds)
    def roc(self):
        if self.data.c!=2: raise ValueError('More than two classes')
        plot_roc(self.actuals, self.probs)
#     def show_best(self, n=8, **kwargs):  self.show_set(self.data.ds2, self.probs_sorted[:n], **kwargs)
#     def show_worst(self, n=8, **kwargs): self.show_set(self.data.ds2, self.probs_sorted[-n:][::-1], **kwargs)
#     def show_set(self, ds, idxs, **kwargs):
#         labels = [f'{ds.labels[i]}:{self.actuals[i]} -> {ds.cls[self.preds[i]]}:{self.preds[i]} (p={self.probs_for_actuals[i]:.4f})'
#                     for i in idxs]
#         ds.show(list(idxs), labels=labels, **kwargs)


class Results_Clsf_Tabular(Results_Clsf):
    def __init__(self, data, df):
        assert 'prob' in df.columns
        assert data.proc.c_dep in df.columns
        self.proc, self.df = data.proc, df
        super(Results_Clsf_Tabular, self).__init__(data, df['prob'].values, df[data.proc.c_dep].values)
        # self.df['prob'] = self.probs if self.data.c>2 else self.probs_bin
    # def corr_cat(self): get_corr(self.df, self.proc.c_cat, plot=True)
    # def corr_cont(self): get_corr(self.df, self.proc.c_cont, plot=True)
    # def corr_prob(self): get_corr(self.df, self.proc.c_cat+self.proc.c_cont+[self.proc.c_dep, 'prob'], c='prob', plot=True)



import time, re, os
import math, random
import numpy as np
import pandas as pd
from functools import partial
import torch
from torch import tensor
import matplotlib.pyplot as plt
import matplotlib as mpl
from dlogicutils.core import listify, camel2snake
from nn2.etc import *


def randn(n): return random.randint(0,n)


class Callback():
    _order=0
    def set_learner(self, learner): self.learner=learner
    def __getattr__(self, k): return getattr(self.learner, k)

    @property
    def name(self):
        name = re.sub(r'Callback$', '', self.__class__.__name__)
        return camel2snake(name or 'callback')

    def __call__(self, cb_name):
        f = getattr(self, cb_name, None)
        if f and f(): return True
        return False


class CancelTrainException(Exception): pass
class CancelEpochException(Exception): pass
class CancelBatchException(Exception): pass


class TestCallback(Callback):
    _order=1
    def after_step(self):
        print(self.iter)
        if self.iter>=5: raise CancelTrainException()


class Recorder(Callback):
    def __init__(self): self.begin_fit()
    def begin_fit(self): self.lrs, self.losses1, self.losses2 = [], [], []

    def after_batch(self):
        if self.train: self.losses1.append(self.loss.detach().cpu())
        else: self.losses2.append(self.loss.detach().cpu())
        if not self.train: return
        if not hasattr(self.optf, 'hypers'): # Pytorch optimizer
            self.lrs.append(self.optf.param_groups[-1]['lr'])
        else: # NN2 optimizer
            self.lrs.append(self.optf.hypers[-1]['lr'])

    def plot_lr(self): plt.figure(); plt.plot(self.lrs); plt.show(); plt.pause(.001)
    def plot_loss1(self): plt.figure(); plt.plot(self.losses1); plt.show(); plt.pause(.001)
    def plot_loss2(self): plt.figure(); plt.plot(self.losses2); plt.show(); plt.pause(.001)


class LR_Find(Callback):
    def __init__(self, max_iter=100, min_lr=1e-6, max_lr=1e1):
        self.max_iter, self.min_lr, self.max_lr = max_iter, min_lr, max_lr
        self.best_loss, self.n = 1e9, 0

    def begin_epoch(self):
        print(f'LR Find: bs={self.data.bs}, n_batches={len(self.data.dl1)}')
        
    def begin_batch(self):
        if not self.train: return
        pos = self.n/self.max_iter
        lr = self.min_lr * (self.max_lr/self.min_lr) ** pos
        self.n += 1
        if not hasattr(self.optf, 'hypers'): # Pytorch optimizer
            for pg in self.optf.param_groups: pg['lr'] = lr
        else: # NN2 optimizer
            for pg in self.optf.hypers: pg['lr'] = lr
            
    def after_step(self):
        if self.n>=self.max_iter or self.loss>self.best_loss*10:
            print('Optimal lr found')
            raise CancelTrainException()
        loss = self.loss.detach().cpu().item()
        if loss < self.best_loss: self.best_loss = loss

    def after_cancel_train(self): self.plot()
    # def after_fit(self): self.plot()

    def plot(self):
        _ = plt.figure(figsize=(5, 3))
        plt.xscale('log')
        plt.plot(self.recorder.lrs[:-1], self.recorder.losses1[:-1])
        plt.show()
        plt.pause(.001)


class BalanceScheduler(Callback):
    def __init__(self, p=1., balance_epochs=None): self.p0, self.p, self.balance_epochs = p, p, balance_epochs
    def begin_epoch(self):
        if not self.train: return
        if self.ep == 0:
            self.data.dl1.set_balance(self.p)
            return
        if self.balance_epochs is None: return
        if self.ep>self.balance_epochs: return
        self.p = self.p0*(1-self.ep/self.balance_epochs)
        self.data.dl1.set_balance(self.p)


class TabMix(Callback):
    def __init__(self, p0=.05, p1=.05): self.p0, self.p1 = p0, p1
    def begin_batch(self):
        if not self.train: return
        n0, n1 = self.x.size(0), self.x.size(1)
        mask0 = torch.randint(n0, (int(self.p0*n0/2),))
        for i in mask0:
            mask1 = torch.randint(n1, (int(self.p1*n1),))
            tmp = self.x[i][mask1].clone()
            self.x[i][mask1] = self.x[n0-i-1][mask1]
            self.x[n0-i-1][mask1] = tmp



import math
import numpy as np
import pandas as pd
import torch
import sklearn.metrics as sk_metrics
from sklearn.metrics import roc_auc_score
import matplotlib.pyplot as plt
import matplotlib as mpl
import seaborn as sns


"""
    Precision:
        A measure of exactness.
        Out of those predicted positive, how many are actual positive.
        Good measure when the costs of False Positive is high (ex: email spam).

    Recall:
        A measure of a classifiers completeness.
        How many of the Actual Positives the model captured.
        Good measure when the costs of False Negatives is high (ex: fraud detection or sick patient detection).

    F1 Score (or F-score)
        A weighted average of precision and recall.
        Seek a balance between Precision and Recall when there is an uneven class (large number of Actual Negatives).
        Higher if both Precision and Recall are higher.

    Receiver Operating Characteristic (ROC)
        ROC curves feature true positive rate on the Y axis, and false positive rate on the X axis.
        Top left corner is the ideal point - a false positive rate of zero, and a true positive rate of one.
        A larger area under the curve (AUC) is better.
        The steepness of ROC curves is  important, to maximize the TPR while minimizing the FPR.

    Suppose a computer program for recognizing dogs in photographs identifies 8 dogs in a picture containing 12 dogs and
    some cats. Of the 8 dogs identified, 5 actually are dogs (true positives), while the rest are cats (false positives).
    The program's recal is 5/12 and the precision is 5/8.

"""


def accuracy(y2, y):
    y2 = torch.max(y2, dim=1)[1]
    return (y2==y).float().mean().item()

def recall(y2, y): # TPR = TP/P (same as sensitivity)
    y2 = torch.max(y2, dim=1)[1]
    tp = torch.mul((y2==y), y.byte()).sum().item()
    p = (y==1).sum().item()
    r = float(tp) / float(p) if p>0 else 0.0
    return r

def sensitivity(y2, y): # TPR = TP/P (same as recall)
    return recall(y2, y)

def specificity(y2, y): # TNR = TN/N
    y2 = torch.max(y2, dim=1)[1]
    tn = torch.mul((y2==y), y.byte()^1).sum().item()
    n = (y==0).sum().item()
    return tn/n if n>0 else 0.0

def precision(y2, y): # TP/(TP+FP)
    y2 = torch.max(y2, dim=1)[1]
    tp = torch.mul((y2==y), y.byte()).sum().item()
    p = (y2==1).sum().item()
    return tp/p if p>0 else 0.0

def accuracy2(y2, y, threshold=0.5):
    y2 = (y2 > threshold).byte()
    y = y.byte()
    return (y2==y).float().mean().item()

def precision2(y2, y, threshold=0.5):
    y2 = (y2 > threshold).byte()
    y = y.byte()
    tp = torch.mul((y2==y), y).sum().item()
    n = (y2==1).sum().item()
    r = float(tp) / float(n) if n>0 else 0.0
    return r

def recall2(y2, y, threshold=0.5):
    y2 = (y2 > threshold).byte()
    y = y.byte()
    tp = torch.mul((y2==y), y).sum().item()
    n = (y==1).sum().item()
    r = float(tp) / float(n) if n>0 else 0.0
    return r

def f1(y2, y):
    p = precision(y2, y)
    r = recall(y2, y)
    if p+r==0.0: return 0.0
    return 2*(p*r)/(p+r)

def f2(y2, y):
    p = precision2(y2, y)
    r = recall2(y2, y)
    return 2*(p*r)/(p+r)

def roc(y2, y): return roc_auc_score(y.cpu().numpy(), np.exp(y2.cpu().numpy()[:,1]))
def gini(y2, y): return 2 * roc(y2, y) - 1
def gini_normalized(y2, y): return gini(y2, y) / gini(y, y)

def gini1(y2, y, cmpcol=0, sortcol=1):
    #y2 = torch.max(y2, dim=1)[1].numpy()
    all = np.asarray(np.c_[ y, y2, np.arange(len(y)) ], dtype=np.float)
    all = all[ np.lexsort((all[:,2], -1*all[:,1])) ]
    totalLosses = all[:,0].sum()
    giniSum = all[:,0].cumsum().sum() / totalLosses
    giniSum -= (len(y) + 1) / 2.
    return giniSum / len(y)

def gini2(actual, pred):
    n = len(actual)
    a_s = actual[np.argsort(pred)]
    a_c = a_s.cumsum()
    giniSum = a_c.sum() / a_c[-1] - (n + 1) / 2.0
    return giniSum / n

def mse1(y2, y):
    return ((y - y2)**2).mean()

def exp_mse(y2, y):
    y = y.exp()
    y2 = y2.exp()
    return ((y - y2)**2).mean()

def exp_rmspe(y2, y):
    y = y.exp()
    y2 = y2.exp()
    r = (y - y2) / y
    return math.sqrt((r**2).mean())

def percent(y2, y):
    return ((y2 - y) / y).mean()

def exp_percent(y2, y):
    return ((y2.exp() - y.exp()).abs() / y.exp()).mean()



import datetime, time, re, os
import numpy as np
import pandas as pd
import pickle
import gc
import torch
import matplotlib.pyplot as plt
from IPython.display import display, clear_output
from dlogicutils.ju import disablePrint, enablePrint, force_print
from nn2.etc import *
from nn2.cbs import Callback
from nn2.metrics import *


def key_metric(learner): return 4+len(learner.metrics)

def metrics_str(metrics):
    if not len(metrics): return ''
    s = ' M: ['
    for v in metrics: s += f'{v:.4f}, '
    s = s[:-2] + ']'
    return s


def log_str(ep, t, loss1, loss2, metrics1, metrics2, new_best, valid=False):
    s = f'Ep: {ep}\tT: {t2s(t)}:{int(t)%60:02d}\tL1: {loss1:.4f} '
    s += metrics_str(metrics1)
    if valid:
        s += f'\tL2: {loss2:.4f} '
        s += metrics_str(metrics2)
    if new_best: s += ' *'
    return s


def printProgressBar(i, n, prefix, w=50):
    p = float(i+1)/n
    filled = int(w*p)
    s = f'\033[31m{"█"*filled}\033[37m{"█"*(w-filled)}\033[0m'
    if p<1.: print(f'\r{prefix} | {s} | {100*p:.1f}%', end='\r')
    else: print(f'\r{" "*(w+len(prefix)+25)}', end='\r')


def plot_init(x_min, x_max):
    plt.ion()
    fig = plt.figure(figsize=(12, 3))
    ax = fig.add_axes([0.2, 0.2, .8, 0.8])
    ax.xaxis.set_ticks(np.arange(x_min, x_max, max(1,int((x_max-x_min))//10)))
    ax.yaxis.set_ticks(np.arange(0., 2.5, .5))
    ax.plot([], [], [], [],)
    return fig, ax


def plot_update(fig, ax, x, y1, y2):
    ax.lines[0].set_xdata(x), ax.lines[0].set_ydata(y1)
    if y2 is not None: ax.lines[1].set_xdata(x), ax.lines[1].set_ydata(y2)
    ax.relim()
    ax.autoscale()
    fig.canvas.draw()


def plot_df(df):
    fig = plt.figure(figsize=(12, 3))
    ax1, ax2 = fig.add_subplot(121), fig.add_subplot(122)
    df.plot(x='Epoch', y=['Loss_1', 'Loss_2'], ax=ax1)
    df.plot(x='Epoch', y=['accuracy_1', 'accuracy_2'], ax=ax2)
    plt.show()
    time.sleep(.01)


class LoggerCallback(Callback):
    
    def __init__(self, mode=None, file_=False, maximize=True, ep_stats=True):
        self.file, self.maximize, self.ep_stats = file_, maximize, ep_stats
        self.mode = ('ipynb' if in_ipynb() else 'console') if mode is None else mode
        if file_:
            fname, dirname = get_log_fname(self.path)
            if not os.path.exists(dirname): os.makedirs(dirname)
            self.logf = open(fname, 'w', 1)
        self.best = 0.0 if self.maximize else 1e9
        self.results = []
        self.y_ep, self.y2_ep, self.stats_1, self.stats_2, self.t0, self.t1 = None, None, None, None, 0, 0
        self.df, self.df_best = None, None
        print(f'Logging to {self.mode}')
        if self.file: print(f'Logging to file: {self.fname}')
        pd.options.display.float_format = '{:,.4f}'.format
        
    def reset_stats(self):
        self.y_ep = torch.tensor([]).long() if self.data.clsf else torch.tensor([]).float()
        self.y2_ep = torch.tensor([]).float()
        self.loss_t, self.n, self.metrics_v = 0., 0, [0.]*len(self.metrics)
        
    def stats(self):
        if self.y_ep is None or self.y2_ep is None: return [0.]*(1+len(self.metrics))
        self.metrics_v = [m(self.y2_ep, self.y_ep) for m in self.metrics]
        return [self.loss_t/self.n] + self.metrics_v
    
    def accumulate(self):
        k = self.x.shape[0]
        loss = self.loss.detach().cpu().item()
        self.loss_t += loss*k
        self.n += k
        self.y_ep = torch.cat((self.y_ep, self.y.detach().cpu()))
        self.y2_ep = torch.cat((self.y2_ep, self.y2.detach().cpu()))

    def update_historic(self):
        row = [self.ep, self.t2] + [self.loss1] + [e for e in self.metrics1] + [self.loss2] + [e for e in self.metrics2]
        self.results.append(row)
        self.df = pd.DataFrame(self.results, columns=self.cols)
       
    def log_epoch(self):
        if self.mode == 'ipynb' and self._plot and self.eps > 1:
            x = self.df['Epoch'].values
            y = self.df[['Loss_1', 'Loss_2']].values
            y0, y1 = y[:,0], y[:,1] if self.data.dl2 is not None else None
            plot_update(self.fig, self.ax, x, y0, y1)
        s = log_str(self.ep, self.t2, self.loss1, self.loss2, self.metrics1, self.metrics2, self.new_best, valid=self.data.dl2 is not None)
        print(s)

    def begin_fit(self):
        if self.data.dl2 is None: self.best = 1e9
        if self.mode == 'ipynb' and self._plot and self.eps > 1: self.fig, self.ax = plot_init(0, self.eps)
        self.cols = ['Epoch', 'Time', 'Loss_1'] + [f'{e.__name__}_1' for e in self.metrics] + ['Loss_2'] + [f'{e.__name__}_2' for e in self.metrics]
        
    def begin_epoch(self):
        self.t0 = time.time()
        self.reset_stats()
        
    def begin_validate(self):
        self.t1 = time.time() - self.t0
        self.stats_1 = [v for v in self.stats()]
        self.reset_stats()
            
    def after_loss(self):
        with torch.no_grad(): self.accumulate()
        
    def after_batch(self):
        printProgressBar(self.iter, self.iters, f'Ep: {self.ep:d} {"train" if self.train else "eval"}:')

    def after_epoch(self):
        self.t2 = time.time() - self.t0 if self.t0 is not None else time.time()
        self.stats_2 = [v for v in self.stats()] if self.data.dl2 is not None else [0]*len(self.stats_1)
        self.loss1, self.loss2 = self.stats_1[0], self.stats_2[0] if self.data.dl2 is not None else 0.
        self.metrics1, self.metrics2 = [v for v in self.stats_1[1:]], [v for v in self.stats_2[1:]]
        loss_metric = not len(self.metrics) or self.data.dl2 is None
        self.new_best = False
        if loss_metric:
            if self.data.dl2 is None:
                self.new_best = self.loss1 < self.best
                if self.new_best: self.best = self.loss1
            else:
                self.new_best = self.loss2 < self.best
                if self.new_best: self.best = self.loss2
        else:
            self.new_best = (self.maximize and self.metrics2[0]>self.best) or (not self.maximize and self.metrics2[0]<self.best)
            if self.new_best: self.best = self.metrics2[0]
        self.update_historic()
        self.log_epoch()
        if self.new_best:
            self.df_best = self.df[self.df['Epoch']==self.ep]
            if self._save: self.save()
        
    def after_fit(self):
        print(f'\nBest: {self.best:.4f}')
        plt.ioff()

    def show_best(self): display(self.df_best)

    def plot_loss1_metric2(self):
        if self.df is None: return
        col_ids = [2, 4+len(self.metrics)]
        z = self.df.iloc[:,col_ids].values
        plot(*list(zip(*z))); plt.show(); time.sleep(.01)


class Runner:
    def __init__(self, path, data, get_learner, nruns=10, eps=10, name='Runner', desc='', load=False):
        self.get_learner, self.data = get_learner, data
        self.nruns, self.eps, self.name, self.desc = nruns, eps, name, desc
        if not Path(f'{path}/runners').exists(): Path(f'{path}/runners').mkdir()
        self.fname = f'{path}/runners/{name}'
        self.learner, self.df, self.best = None, None, 0.
        if load: self.load()

    def __call__(self, mute=True):
        self.best, self.best_file = 0., None
        if mute: disablePrint()
        for i in range(self.nruns):
            force_print(f'\rRun {i}')
            learner = self.get_learner(self.data, name= f'{self.name}_{i}', plot=False)
            self.learner = learner
            learner.fit(self.eps)
            learner.load(mute=True)
            df1 = learner.logger.df_best
            df1['Best'] = ''
            self.df = df1.copy() if self.df is None else pd.concat((self.df,df1),ignore_index=True)
            if learner.logger.best > self.best:
                self.best, self.best_file = learner.logger.best, f'{self.name}_{i}'
                print(f'New best: {self.best:.4f}')
                display(self.df.iloc[[-1]])
            self.df['Best'] = ''
            self.df.loc[self.df.iloc[:,key_metric(learner)]==self.best, 'Best'] = '*'
            self.save()
            learner.close()
            del learner
            gc.collect()
        self.save()
        print(f'\nBest file: {self.best_file}')
        display(self.df)
        enablePrint()

    def save(self):
        meta = {'desc': self.desc, 'best': self.best, 'best_file': self.best_file}
        pickle.dump(meta, open(self.fname+'.pickle', 'wb'))
        self.df.to_csv(self.fname+'.csv', index=False)

    def load(self):
        p_pickle, p_csv = Path(f'{self.fname}.pickle'), Path(f'{self.fname}.csv')
        if not p_pickle.exists() or not p_csv.exists(): return
        meta = pickle.load(open(self.fname+'.pickle', 'rb'))
        if not 'cols' in meta or not 'best' in meta or not 'best_file' in meta: return
        print(f'Loaded from file {self.fname}: {meta}')
        self.desc, self.best, self.best_file = meta['desc'], meta['best'], meta['best_file']
        self.df = pd.read_csv(self.fname+'.csv')
        self.df['Best'] = self.df['Best'].fillna('')

    def df_best(self):
        if self.learner is None: return None
        return self.df[self.df.iloc[:,key_metric(self.learner)]==self.best].iloc[[0]]
